import os
from collections import defaultdict, OrderedDict

"""
The keys of the following dictionaries refer to specific emotion codenames
    used in each database.
Each entry is a list of aliases which can be used to refer to
    the corresponding emotion during database initialization.
"""
IEMOCAP_EMO_CODES = {'neu': ['neu', 'neutral'],
                     'hap': ['hap', 'happy', 'happiness'],
                     'sad': ['sad', 'sadness'],
                     'ang': ['ang', 'angry', 'anger'],
                     'sur': ['sur', 'surprise', 'surprised'],
                     'fea': ['fea', 'fear'],
                     'dis': ['dis', 'disgust', 'disgusted'],
                     'fru': ['fru', 'frustrated', 'frustration'],
                     'exc': ['exc', 'excited', 'excitement'],
                     'oth': ['oth', 'other', 'others']}

EMODB_EMO_CODES =   {'N': ['neu', 'neutral'],
                     'F': ['hap', 'happy', 'happiness'],
                     'T': ['sad', 'sadness'],
                     'W': ['angry','anger'],
                     'L': ['bored', 'boredom'],
                     'A': ['fea', 'fear'],
                     'E': ['dis', 'disgust', 'disgusted']}
SAVEE_CODES =        {'SN': ['neu', 'neutral'],
                     'SF': ['hap', 'happy', 'happiness'],
                     'ST': ['sad', 'sadness'],
                     'SW': ['angry','anger'],
                     'Sfru':['fru', 'frustrated', 'frustration'],
                     'Ssur':['sur', 'surprise', 'surprised'],
                     'SE': ['dis', 'disgust', 'disgusted']}
RAVDESS_CODES =      {'RN': ['neu', 'neutral'],
                     'RF': ['hap', 'happy', 'happiness'],
                     'RT': ['sad', 'sadness'],
                     'RW': ['angry','anger'],
                     'Rsur': ['sur', 'surprise', 'surprised'],
                     'RA': ['fea', 'fear'],
                     'RC': ['cal', 'calm', 'calm'],
                     'RE': ['dis', 'disgust', 'disgusted']}


class IEMOCAP_database():
    def __init__(self, database_dir, emot_map = {'ang': 0, 'sad':1, 'hap':2, 'neu':3},
                        include_scripted=False):

        #Path
        self.database_dir = database_dir

        #Emotion to label mapping for features
        self.emot_map = emot_map

        #IEMOCAP Session name
        self.sessions = ['Session1','Session2','Session3','Session4','Session5']

        #IEMOCAP available emotion classes
        self.all_emo_classes = IEMOCAP_EMO_CODES.keys()

        #to include scripted session
        self.include_scripted = include_scripted

    def get_speaker_id(self, session, gender):

        return session[-1]+gender

    def get_classes(self):

        classes={}
        for key,value in self.emot_map.items():
            if value in classes.keys():
                classes[value] += '+'+key
            else:
                classes[value] = key

        return classes

    def get_files(self):
        emotions = self.emot_map.keys()
        dataset_dir = self.database_dir
        all_speaker_files = defaultdict()
        total_num_files = 0
        for session_name in os.listdir(dataset_dir):

            if session_name not in self.sessions:
                continue
            wav_dir = os.path.join(dataset_dir, session_name, "sentences/wav")
            lab_dir = os.path.join(dataset_dir, session_name, "dialog/EmoEvaluation")

            M_wav, F_wav = list(), list()
            for conversation_folder in os.listdir(wav_dir):
                #omit hidden folders
                if conversation_folder.startswith('.'):
                    continue

                if self.include_scripted == False:
                    # Only use improvised data, for example ".../wav/Ses01F_impro01"
                    if conversation_folder[7:12] != "impro":
                        continue

                # Path to the directory containing all the *.wav files of the
                # current conversation
                conversation_dir = os.path.join(wav_dir, conversation_folder)

                # Get labels of all utterance in the current conversation
                label_path = os.path.join(lab_dir, conversation_folder + ".txt")
                labels = dict()
                with open(label_path, "r") as fin:
                    #print(label_path)
                    for line in fin:
                        # If this line is sth like
                        # [6.2901 - 8.2357]	Ses01F_impro01_F000	neu	[2.5000, 2.5000, 2.5000]
                        if line[0] == "[":
                            t = line.split()
                            # For e.g., {"Ses01F_impro01_F000": "neu", ...}
                            labels[t[3]] = t[4]

                # Get a list of paths to all *.wav files
                wav_files = []
                for wav_name in os.listdir(conversation_dir):
                    #omit hidden folders
                    if wav_name.startswith('.'):
                        continue
                    #omit non .wav files
                    name, ext = os.path.splitext(wav_name)
                    if ext != ".wav":
                        continue
                    #emotion label
                    emotion = labels[name]
                    if emotion not in emotions:
                        continue
                    label = self.emot_map[emotion]

                    wav_files.append((os.path.join(conversation_dir, wav_name), label))

                #separate into individual speakers
                F_wav.extend([emowav for emowav in wav_files if emowav[0][-8] == "F"])
                #F_lab.append(label_path)
                M_wav.extend([emowav for emowav in wav_files if emowav[0][-8] == "M"])
                #M_lab.append(label_path)

            #Put speaker utterance and label paths into dictionary
            all_speaker_files[self.get_speaker_id(session_name,'M')] = M_wav
            all_speaker_files[self.get_speaker_id(session_name,'F')] = F_wav

            total_num_files += len(M_wav) + len(F_wav)
        print(f'\nNUMBER OF FILES: {total_num_files}\n')
        return all_speaker_files


class EMODB_database():

    def __init__(self, database_dir, emot_map={'neutral':0,'happy':1,'sad':2,'angry':3,'bored':4,'fear':5,'disgust':6}):

        #Path
        self.database_dir = database_dir

        #Emotion to label mapping
        self.emot_map = emot_map

        self.all_emo_codes = EMODB_EMO_CODES


    def _get_speaker_id(self,speaker_code):
        emodb_speaker_ids = {'03':'1M','08':'1F','09':'2F','10':'2M','11':'3M',
                             '12':'4M','13':'3F','14':'4F','15':'5M','16':'5F'}

        return emodb_speaker_ids[speaker_code]

    def get_classes(self):

        classes={}
        for key,value in self.emot_map.items():
            if value in classes.keys():
                classes[value] += '+'+key
            else:
                classes[value] = key

        return classes


    def get_files(self):

        all_files =defaultdict(list)

        #Get all filenames
        filenames = []
        for (dirpath, dirnames, fnames) in os.walk(self.database_dir):
            for name in fnames:
                if name.startswith('.'):
                    continue
                filenames.append(name)

        assert len(filenames) == 535

        emotion_codes_to_labels = {}
        #Map the required emotion codes to class label
        for clss in self.emot_map.keys():
            for emo_code, emo_class in self.all_emo_codes.items():
                if clss in emo_class:
                    emotion_codes_to_labels[emo_code] = self.emot_map[clss]

        #Grouped the files based on speaker, into dictionary
        for fname in filenames:
            speaker_code = fname[0:2]
            emotion_code = fname[5]

            speaker_id = self._get_speaker_id(speaker_code)
            emotion_label = emotion_codes_to_labels[emotion_code]

            all_files[speaker_id].append((dirpath+fname,emotion_label))

        return all_files


class RAVDESS_database():
    def __init__(self, database_dir, emot_map={'neutral':0,'happy':1,'sad':2,'angry':3,'surprise':4,'calm':5,'disgust':6}):

        #Path
        self.database_dir = database_dir

        #Emotion to label mapping
        self.emot_map = emot_map

        self.all_emo_codes = EMODB_EMO_CODES


    def _get_speaker_id(self,speaker_code):

        return ravdess_speaker_ids[speaker_code]

    def get_classes(self):

        classes={}
        for key,value in self.emot_map.items():
            if value in classes.keys():
                classes[value] += '+'+key
            else:
                classes[value] = key

        return classes


    def get_files(self):

        all_files =defaultdict(list)

        #Get all filenames
        filenames = []
        for (dirpath, dirnames, fnames) in os.walk(self.database_dir):
            for name in fnames:
                if name.startswith('.'):
                    continue
                filenames.append(name)

        assert len(filenames) == 7356

        emotion_codes_to_labels = {}
        #Map the required emotion codes to class label
        for clss in self.emot_map.keys():
            for emo_code, emo_class in self.all_emo_codes.items():
                if clss in emo_class:
                    emotion_codes_to_labels[emo_code] = self.emot_map[clss]

        #Grouped the files based on speaker, into dictionary
        for fname in filenames:
            speaker_code = fname[0:2]
            emotion_code = fname[5]

            speaker_id = self._get_speaker_id(speaker_code)
            emotion_label = emotion_codes_to_labels[emotion_code]

            all_files[speaker_id].append((dirpath+fname,emotion_label))

        return all_files


class SAVEE_database():



    def __init__(self, database_dir, emot_map={'neutral':0,'happy':1,'sad':2,'angry':3,'surprise':4,'frustrated':5,'disgust':6}):

        #Path
        self.database_dir = database_dir

        #Emotion to label mapping
        self.emot_map = emot_map

        self.all_emo_codes = EMODB_EMO_CODES


    def _get_speaker_id(self,speaker_code):

        return savee_speaker_ids[speaker_code]

    def get_classes(self):

        classes={}
        for key,value in self.emot_map.items():
            if value in classes.keys():
                classes[value] += '+'+key
            else:
                classes[value] = key

        return classes


    def get_files(self):

        all_files =defaultdict(list)

        #Get all filenames
        filenames = []
        for (dirpath, dirnames, fnames) in os.walk(self.database_dir):
            for name in fnames:
                if name.startswith('.'):
                    continue
                filenames.append(name)

        assert len(filenames) == 480I 

        emotion_codes_to_labels = {}
        #Map the required emotion codes to class label
        for clss in self.emot_map.keys():
            for emo_code, emo_class in self.all_emo_codes.items():
                if clss in emo_class:
                    emotion_codes_to_labels[emo_code] = self.emot_map[clss]

        #Grouped the files based on speaker, into dictionary
        for fname in filenames:
            speaker_code = fname[0:2]
            emotion_code = fname[5]

            speaker_id = self._get_speaker_id(speaker_code)
            emotion_label = emotion_codes_to_labels[emotion_code]

            all_files[speaker_id].append((dirpath+fname,emotion_label))

        return all_files
SER_DATABASES = {'IEMOCAP': IEMOCAP_database,
                 'EMODB'  : EMODB_database,
                  'SAVEE'  :SAVEE_database,
                 'RAVDESS'  : RAVDESS_database}
